#!/usr/bin/env perl

######################################################################
# blikII is a continuation of blik for FICS                          #
# Copyright (C) 2020  Asher Gordon <AsDaGo@posteo.net>               #
#                                                                    #
# This file is part of blikII.                                       #
#                                                                    #
# blikII is free software: you can redistribute it and/or modify it  #
# under the terms of the GNU Affero General Public License as        #
# published by the Free Software Foundation, either version 3 of the #
# License, or (at your option) any later version.                    #
#                                                                    #
# blikII is distributed in the hope that it will be useful, but      #
# WITHOUT ANY WARRANTY; without even the implied warranty of         #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  #
# Affero General Public License for more details.                    #
#                                                                    #
# You should have received a copy of the GNU Affero General Public   #
# License along with blikII.  If not, see                            #
# <https://www.gnu.org/licenses/>.                                   #
######################################################################

use strict;
use warnings;
use Net::Telnet;
use Time::HiRes qw(time sleep);
use Time::Duration 'duration';
use Time::ParseDate;
use Getopt::Long qw(GetOptionsFromArray :config bundling);
use File::HomeDir ();
use Text::Wrap;
use IPC::Open2;
use FindBin;
use POSIX 'strftime';
use v5.16;		# For fc
use subs qw(tell log);	# Override CORE functions with Main functions

my $name = 'blikII';
my $password;
my $version = '0.1';
my $operator = 'AsDaGo';

my $host = 'freechess.org';
my $port = 5000;

my $prompt = 'fics% ';

# A pattern that will match a handle on the server.
my $handle_pat = qr/[[:alpha:]]+/;

# The maximum length of a 'tell' and 'qtell'.
my $max_tell = 1000;
my $max_qtell = 1000;

# The maximum length of a fortune.
my $max_fortune = 400;

# The time (in seconds) that tells will be queued by the server.
my $tell_queue_time;

# The time (in seconds) that we need to do something in order to
# prevent the server from kicking us off.
my $ping_time;

# The time of the last tell.
my $last_tell_time;

# The age a tell needs to be (in seconds) before it is no longer
# considered as part of the quota.
my $quota_age = 30;

# The size of the tell quota.
my $quota_size = 3;

# How bored we are.
my $boredom = 0;

# How long it takes us (in seconds) to get more bored.
my $patience = 3600;

# Whether we should update boredom and the last tell time.
my $update_boredom = 1;
my $update_last_tell = 1;

# Messages we say depending on how bored we are. After these, we just
# say "I'm so bored.", "I'm so, so bored." and so on.
my @bored_messages = split "\n", <<EOF;
I'm bored.
I'm still bored.
I'm bored out of my skull.
Someone talk to me!
Where is everyone!?!
EOF

# Whether we should reverse tells/shouts.
my $reverse_tells = 0;

# The channels we will listen to.
my @channels = (37, 38, 50, 53, 88, 120);

# An environment variable used when we are restarted by a user. This
# is either the handle of a user or the number of a channel or the
# empty string for a shout.
my $restart_target_var = 'BLIKII_RESTART_TARGET';

# The location of old blik phrases.
my $blik_phrase_file = "$FindBin::Bin/blik_phrases";

# Where we will store and read configuration files.
my $config_dir;
{
    my $home = File::HomeDir->my_home;
    unless (defined $home) {
	warn 'No home directory; using current directory';
	$home = '.';
    }
    $config_dir = "$home/.blikii";
}
my $vars_dir = "$config_dir/vars";
my $password_file = "$config_dir/password";

my $usage = <<EOF;
Usage: $0 [OPTION]... [HOST [PORT]]
Connect to the chess server at HOST (default $host) through port PORT
(default $port) and start running.

  -n, --name=NAME         use NAME to log in rather than $name
  -v, --verbose           print verbose output
  -l, --log-file=FILE     log user communications to FILE
      --debug-file=FILE   log all I/O to FILE (NOTE: this does not
                            respect users' 'log' settings)
  -h, --help              print this help and exit
  -V, --version           print version information and exit
EOF

# Remember how we were run in case we need to restart ourself.
my @cmdline = ($0, @ARGV);

GetOptions
    'n|name=s'		=> \$name,
    'l|log-file=s'	=> \my $log_filename,
    'debug-file=s'	=> \my $debug_filename,
    'v|verbose'		=> \my $verbose,
    'h|help'		=> sub { print $usage; exit },
    'V|version'		=> sub { print "$name $version\n"; exit },
    or die $usage;

# After we got the options, get rid of the 'permute' option. We don't
# want to permute for our commands, because that can be confusing
# (e.g., 'tell $name figlet The -c option centers text.' does not do
# what you think).
Getopt::Long::Configure 'require_order';

$host = shift if @ARGV;
$port = shift if @ARGV;
die $usage if @ARGV;

die "Invalid handle: $name" unless $name =~ /^$handle_pat$/;

unless (defined $password) {
    # Get the password from the password file.
    if (open my $file, '<', $password_file) {
	$password = <$file>;
	close $file or warn "Cannot close $password_file: $!";
    }
    else {
	warn "Cannot open $password_file: $!";
    }
}

sub create_dir {
    foreach (@_) {
	unless (-d) {
	    die "$_ exists but is not a directory" if -e;
	    mkdir or die "Cannot create $_: $!";
	}
    }
}

# Create the configuration directory and variables directory if they
# don't already exist.
create_dir $config_dir, $vars_dir;

# Get the defaults for figlet.
my $figlet_font_dir = `figlet -I2`;
unless (defined $figlet_font_dir) {
    $figlet_font_dir = '/usr/share/figlet';
    warn 'Cannot get figlet font directory; ' .
	"defaulting to $figlet_font_dir";
}
chomp $figlet_font_dir;
my $figlet_default_font = `figlet -I3`;
unless (defined $figlet_default_font) {
    $figlet_default_font = 'standard';
    warn 'Cannot get default figlet font; ' .
	"defaulting to $figlet_default_font";
}
chomp $figlet_default_font;
my @figlet_font_extensions;
{
    my $formats = `figlet -I5`;
    unless (defined $formats) {
	$formats = 'flf2 tlf2';
	warn 'Cannot get supported figlet font formats; ' .
	    "defaulting to '$formats'";
    }
    foreach (split ' ', $formats) {
	s/\d+$//;
	push @figlet_font_extensions, $_;
    }
}

# Get the available figlet fonts.
my @figlet_fonts;
while (<$figlet_font_dir/*>) {
    next unless m|^.*?([^/]+)\.([^/.]+)$|;
    my ($font, $ext) = ($1, $2);
    my $found = 0;
    foreach (@figlet_font_extensions) {
	if ($ext eq $_) {
	    $found = 1;
	    last;
	}
    }
    next unless $found;
    push @figlet_fonts, $font;
}

# Get the available cowfiles for cowsay.
my @cowsay_cowfiles;
{
    my @cowfiles = `cowsay -l` or warn 'Cannot get default cowfiles';
    foreach (@cowfiles) {
	next if /^Cow files in .+:$/;
	push @cowsay_cowfiles, split;
    }
}
@cowsay_cowfiles = sort @cowsay_cowfiles;

# The text filter programs. There's not really a good way to get them
# programatically, since they are separate programs. Each key is the
# name of the program, and each value is a short description of the
# program. Note that we do not include 'unicode' here, because that
# cannot be transferred across FICS using tells. We don't have to
# handle 'jibberish' specially, because that does not include unicode
# anyway.
my %filter_progs = (
    b1ff	=> 'the B1FF filter',
    LOLCAT	=> 'as seen in internet gifs everywhere',
    cockney	=> 'Cockney English',
    chef	=> 'convert English to Mock Swedish',
    eleet	=> 'k3wl hacker slang',
    fanboy	=> 'speak like a fanboy',
    fudd	=> 'elmer Fudd',
    jethro	=> 'hillbilly text filter',
    jive	=> 'jive English',
    jibberish	=> 'a random selection of the rest of the filters',
    ken		=> 'English into Cockney, computer-related rhyming slang',
    kraut	=> 'generates text with a bad German accent',
    kenny	=> 'generates text as spoken by Kenny on South Park',
    ky00te	=> 'a very cute (and familiar to FurryMuck fans) accent',
    nethackify	=> 'wiped out text like can be found in nethack',
    newspeak	=> 'a-la-1984',
    censor	=> 'cda-ize text',
    nyc		=> 'brooklyn English',
    pirate	=> 'talk like a pirate',
    rasterman	=> 'talk like Carsten Haitzler',
    scottish	=> 'fake scottish (dwarven) accent filter',
    spammer	=> 'turns honest text into spam',
    scramble	=> 'scramble the "inner" letters',
    studly	=> 'studly caps',
    'upside-down' => 'flips text upside down',
    pig		=> 'Pig Latin'
);
my @filter_progs = sort keys %filter_progs;

# Help for all the filters.
my $filters_help = '';
{
    my $maxlen;
    foreach (@filter_progs) {
	my $len = length;
	$maxlen = $len if ! defined $maxlen || $len > $maxlen;
    }
    $filters_help .= "  $_   " . (' ' x ($maxlen - length)) .
	$filter_progs{$_} . "\n" foreach @filter_progs;
}
# So it can be used more naturally in a here-doc.
chomp $filters_help;

# Whether process_message should run the message through a filter even
# in list context.
my $allow_filter = 0;

# Variable specifications. Each key is the name of the variable and
# each value is a hash reference containing its type, default value,
# acceptable values (or empty for any value), and other
# information. The acceptable values list must explicitly contain the
# empty string if that is allowed.
my @variables = (
    log => {
	type	=> 'bool',
	default	=> 1,
	help	=>
	    <<EOF
Whether $name should log your communications.

Logging helps $operator improve $name, but you can set this variable to 0 if you'd rather not be logged.
EOF
    },
    delay => {
	type	=> 'float',
	default	=> 1,
	range	=> [0, 10],
	help	=>
	    <<EOF
How long (in seconds) $name should wait before responding.

This does not apply to command output.
EOF
    },
    'notify-quota' => {
	type	=> 'bool',
	default	=> 1,
	help	=>
	    <<EOF
Whether you will be notified when exceeding quota.

If this variable is true, you will be notified when you attempt to exceed the response quota for shouts or any channel tell. Otherwise, if this variable is false, you will be silently ignored.
EOF
    },
    filter => {
	type	=> 'string',
	default	=> '',
	values	=> ['', @filter_progs],
	help	=>
	    <<EOF
A filter to run all output from $name through.

To run the output through no filter, unset this variable using 'tell $name set filter'.

Available filters are:

$filters_help

See also the 'filter' command.
EOF
    }
);

# Change @variables to the keys only, in the order they were
# given. %variables will be the hash.
my %variables = @variables;
for (my $i = 1; $i < @variables; $i++) {
    splice @variables, $i, 1;
}

# The variables for each user.
my %user_vars;

# Convert a string to a number, or return undef if the string is not a
# valid number.
sub to_num {
    my $num = @_ ? $_[0] : $_;
    return eval {
	delete local $SIG{__DIE__};
	local $SIG{__WARN__} = sub { die };
	$num + 0;
    };
}

# Return a hash with the default values for variables.
sub get_default_vars {
    map { $_ => $variables{$_}->{default} } keys %variables;
}

# Check whether the type given by the first argument is a numeric
# type.
sub is_numeric_type {
    my $type = @_ ? $_[0] : $_;
    foreach (qw(bool int float)) {
	return 1 if $type eq $_;
    }
    return 0;
}

# Check if a value is valid for a variable.
sub var_is_valid {
    my ($var, $value) = @_;
    my %spec = %{$variables{$var}};
    my $type = $spec{type};
    my @values;
    @values = @{$spec{values}} if exists $spec{values};
    my $numeric = is_numeric_type $type;

    if ($numeric) {
	$value = to_num $value;
	return 0 unless defined $value;
	return 0 if $type eq 'bool' && $value != 0 && $value != 1;
	return 0 if $type eq 'int' && $value != int $value;

	if (exists $spec{range}) {
	    my ($lower, $upper) = @{$spec{range}};
	    return 0 if $value < $lower || $value > $upper;
	}
    }

    return 1 unless @values;
    foreach (@values) {
	if ($numeric) {
	    return 1 if $value == $_;
	}
	else {
	    return 1 if $value eq $_;
	}
    }

    return 0;
}

# Load the user variables.
while (<$vars_dir/*>) {
    unless (m|^\Q$vars_dir/\E($handle_pat)$|) {
	warn "Invalid user variable file: $_";
	next;
    }

    my $filename = $_;
    my $handle = fc $1;
    my $file;
    unless (open $file, '<', $filename) {
	warn "Cannot open $filename: $!";
	next;
    }

    my %vars = get_default_vars;
    while (<$file>) {
	chomp;
	unless (/^\s*(.+?)\s*=\s*(.*?)\s*$/) {
	    warn "$filename:$.: invalid line: $_\n";
	    next;
	}
	my ($key, $value) = ($1, $2);
	unless ($key =~ /^[[:alnum:]\-_]+$/) {
	    warn "$filename:$.: invalid variable name: $key\n";
	    next;
	}
	unless (var_is_valid $key, $value) {
	    warn "$filename:$.: invalid variable value: $value\n";
	    next;
	}
	$value += 0 if is_numeric_type $variables{$key}->{type};
	$vars{$key} = $value;
    }

    close $file or warn "Cannot close $filename: $_";

    $user_vars{$handle} = \%vars;
}

# Check if the users given by the first and second arguments are, in
# fact, the same users.
sub user_eq {
    fc $_[0] eq fc $_[1];
}

# Return variables for the user specified by the first argument (or $_
# if omitted). If more than one argument is given, return a list of
# the values of variables as specified by each of the arguments after
# the first argument.
sub get_vars {
    my $user = fc (@_ ? shift : $_);
    my %vars = exists $user_vars{$user} ?
	%{$user_vars{$user}} : get_default_vars;
    return %vars unless @_;
    return map {
	die "Invalid variable: $_" unless exists $vars{$_};
	my $value = $vars{$_};
	die "Invalid value: $value" unless var_is_valid $_, $value;
	$value;
    } @_;
}

# Set the variables for the user specified by the first argument to
# the variables and values specified by the rest. Then, save the
# variable file. Returns true on success or false on error.
sub set_vars {
    die 'No args given' unless @_;
    my $user = fc shift;
    my %new_vars = @_;
    my %vars = get_vars $user;

    foreach my $var (keys %new_vars) {
	die "Invalid variable: $var"
	    unless exists $variables{$var};
	my $value = $new_vars{$var};
	die "Invalid value: $value" unless var_is_valid $var, $value;
	$vars{$var} = $value;
    }
    $user_vars{$user} = \%vars;

    # Now save the variables.
    my $filename;
    while (<$vars_dir/*>) {
	m|^\Q$vars_dir/\E(.+)$| or die "Invalid file: $_";
	if (user_eq $user, $1) {
	    $filename = $_;
	    last;
	}
    }
    $filename = "$vars_dir/$user" unless defined $filename;

    my $file;
    unless (open $file, '>', $filename) {
	warn "Cannot open $filename: $_";
	return 0;
    }

    foreach (@variables) {
	unless (print $file "$_=$vars{$_}\n") {
	    warn "Cannot write to $filename: $_";
	    close $file;
	    return 0;
	}
    }

    unless (close $file) {
	warn "Cannot close $filename: $_";
	return 0;
    }

    return 1;
}

sub backup_and_open {
    my $filename = $_[0];
    return undef unless defined $$filename;

    if (-e $$filename) {
	my $new = "$$filename~";
	rename $$filename, $new
	    or warn "Cannot rename $$filename to $new: $!";
    }

    my $file;
    return $file if open $file, '>', $$filename;
    warn "Cannot open $$filename: $!";
    undef $$filename;
    return undef;
}

my $log_file = backup_and_open \$log_filename;
my $debug_file = backup_and_open \$debug_filename;

# Used to check if we should log communications from a user.
my $current_user;

# Log the given arguments to standard output if $verbose, and
# $log_file if it is defined.
sub log {
    if (defined $current_user) {
	my ($log) = get_vars $current_user, 'log';
	return 0 unless $log;
	unshift @_, "$current_user:";
    }
    my $now = strftime '%Y-%m-%d %H:%M:%S', localtime;
    my $line = "$now: @_";
    my $ret = 1;
    if ($verbose) {
	$ret = (print $line) && $ret;
	STDOUT->flush;
    }
    if (defined $log_file) {
	$ret = (print $log_file $line) && $ret;
	$log_file->flush;
    }
    return $ret;
}

my $telnet;

# Die unless the error was a timeout.
sub die_unless_timeout {
    die $_[0] unless $telnet->timed_out;
}

# Get a line and remove the "\r" at the beginning (if
# present). Returns the line on success, or undef on error.
sub getline {
    my $line = $telnet->getline(@_);
    return undef unless defined $line;
    chomp $line;
    $line =~ s/^\r//;
    return $line;
}

# Run a command through $telnet and return the output. If $verbose is
# true, also print the command.
sub cmd {
    my $cmd = "@_";
    log "> $cmd\n";

    # $telnet->cmd doesn't get the output correctly for some reason,
    # so we have to do it manually.
    return unless $telnet->print($cmd);
    my $first = getline Errmode => \&die_unless_timeout, Timeout => 3;
    return unless defined $first;
    $first =~ s/^$prompt//;
    my ($prematch) =
	$telnet->waitfor(Match => '/\r'.$prompt.'$/',
			 Errmode => \&die_unless_timeout,
			 Timeout => 3);
    return unless defined $prematch;
    my @output = ($first, (split "\n", $prematch));
    foreach (@output) {
	s/^\r//;
	log "< $_\n";
    }
    return wantarray ? @output ? @output : '' : 1;
}

sub reset_tell {
    $last_tell_time = time if $update_last_tell;
    $boredom = 0 if $update_boredom;
}

# Retern a message depending on how bored we are.
sub bored_message {
    my $bored = @_ ? shift : $boredom;
    my @msgs = @_ ? @_ : @bored_messages;
    if ($bored > $#msgs) {
	return "I'm " . (join ', ', (('so') x ($bored - $#msgs))) .
	    ' bored!!!';
    }
    else {
	return $msgs[$bored];
    }
}

# Run text through a filter. Returns the filtered text on success or
# undef on error.
sub do_filter {
    my ($filter, $warn, $text) = @_;

    unless (exists $filter_progs{$filter}) {
	$warn->("Invalid filter: $filter");
	return undef;
    }

    my $failed = 0;
    my $pid;
    my $filter_out;
    my $filter_in;
    local $SIG{PIPE} = sub { $failed = 1 };
    $failed = 1 unless defined eval {
	$pid = open2 $filter_out, $filter_in, $filter;
    };
    if ($failed) {
	$warn->("Failed to execute $filter; " .
		"please contact $operator");
	return undef;
    }

    print $filter_in "$text\n";
    close $filter_in
	or warn "Cannot close write end of $filter pipe";
    local $/;
    my $output = <$filter_out>;
    close $filter_out
	or warn "Cannot close read end of $filter pipe";
    $output =~ s/\s+$//s;

    waitpid $pid, 0;
    my $status = $?;
    if ($status) {
	$warn->("$filter failed; please contact $operator");
	return undef;
    }

    return $output;
}

# Process a message and return the result.
sub process_message {
    die 'Called in scalar context with multiple arguments'
	if ! wantarray && @_ > 1;

    # Don't run it through a filter if $allow_filter is false.
    my $filter = '';
    ($filter) = get_vars $current_user, 'filter'
	if defined $current_user && $allow_filter;

    if ($filter) {
	my $output = do_filter $filter,
	    sub { warn @_ }, (join "\n", @_);
	if (defined $output) {
	    @_ = split "\n", $output;
	    @_ = ("@_") unless wantarray;
	}
    }

    if ($reverse_tells) {
	s/^(\s*)(.*?)(\s*)$/$1 . (reverse $2) . $3/e foreach @_;
    }

    return wantarray ? @_ : "@_";
}

# The subroutine queue.
my @queue;

# Queue a subroutine to be called at a certain time. The first
# argument is the time at which the subroutine should be called, the
# second argument is the user that caused the subroutine to be called
# (may be undef), the third argument is the subroutine to be called,
# and the rest are arguments to pass to the subroutine.
sub queue_absolute {
    die 'Wrong number of arguments' if @_ < 4;
    my ($when, $user, $filter, $func, @args) = @_;
    die "$func is not a CODE ref" unless ref $func eq 'CODE';
    push @queue, [$when, $user, $filter, $func, @args];
}

# Like queue_absolute, but take a relative time rather than an
# absolute time.
sub queue {
    my $delay = shift;
    queue_absolute $delay + time, @_;
}

# Queue a subroutine using the 'delay' variable of the given user to
# determine when the subroutine should be run.
sub queue_for_user {
    my $user = shift;
    die 'User is not defined' unless defined $user;
    my ($delay) = get_vars $user, 'delay';
    queue $delay, $user, $allow_filter, @_;
}

# Run all queued subroutines, using the subroutine given by the first
# argument to determine if a queued sub is ready to run.
sub run_queue_if_ready {
    my $ready = shift;
    my @q;
    for (my $i = 0; $i < @queue;) {
	my @item = @{$queue[$i]};
	my ($when, $user, $filter, $func, @args) = @item;
	if ($ready->(@item)) {
	    splice @queue, $i, 1;
	    push @q, [$when, $user, $filter, $func, @args];
	}
	else {
	    $i++;
	}
    }

    # Sort the subroutines queued to run, and then run them in order.
    @q = sort { $a->[0] <=> $b->[0] } @q;
    foreach (@q) {
	my (undef, $user, $filter, $func, @args) = @$_;
	my $prev_user = $current_user;
	my $prev_allow_filter = $allow_filter;
	$current_user = $user;
	$allow_filter = $filter;
	$func->(@args);
	$current_user = $prev_user;
	$allow_filter = $prev_allow_filter;
    }
}

# Run all queued subroutines if it is time for them to be run.
sub run_queue {
    my $now = @_ ? $_[0] : time;
    run_queue_if_ready sub { $now >= $_[0] };
}

# Run all queued subroutines belonging to the given user, whether or
# not they are queued to run. If no user is given, run all tells.
sub flush_queue {
    my $ready;
    if (@_) {
	my $user = $_[0];
	$ready = sub { defined $_[1] && user_eq $_[1], $user };
    }
    else {
	$ready = sub { 1 };
    }
    run_queue_if_ready $ready;
}

# Quotas. These are implemented on a per-channel/shout basis.
my %quota;

# Get the quota array ref for the given user and channel. If called in
# a list context, returns both the quota array ref and the current
# time or the time specified by the third argument if given.
sub get_quota {
    my ($user, $channel, $now) = @_;

    # Ignore quota for direct tells.
    return unless defined $channel;
    $channel = 'shout' if $channel eq 'it';

    $quota{$user} = {} unless exists $quota{$user};
    $quota{$user}->{$channel} = []
	unless exists $quota{$user}->{$channel};
    my $quota = $quota{$user}->{$channel};

    # Expire sufficiently old items.
    for (my $i = 0; $i < @$quota;) {
	if ($quota->[$i] + $quota_age <= $now) {
	    splice @$quota, $i, 1;
	}
	else {
	    $i++;
	}
    }

    return $quota unless wantarray;
    $now = time unless defined $now;
    return $quota, $now;
}

# Check if a user is over quota.
sub over_quota {
    my ($quota, $now) = get_quota @_;
    return 0 unless defined $quota;
    return @$quota >= $quota_size;
}

# Add an item to the quota.
sub add_quota {
    my ($user) = @_;
    my ($quota, $now) = get_quota @_;
    return unless defined $quota;
    die "$user is already over quota" unless @$quota < $quota_size;
    push @$quota, $now if defined $quota;
}

# Check how long until the sufficient quota items expire.
sub next_quota_availability {
    my ($user) = @_;
    my ($quota, $now) = get_quota @_;

    # We should only be called when the user is over quota.
    die "No quota for $user" unless defined $quota;

    my @quota = @$quota;
    die "$user is not over quota" if @quota < $quota_size;

    # Get the time at which sufficient quota items will have expired.
    @quota = sort { $a <=> $b } @quota;
    return $quota[$quota_size - @quota] + $quota_age - $now;
}

# Perform a tell while making sure it won't be queued. If the first
# argument is undef, shout instead of telling. If the first argument
# is a negative integer, use 'it' instead of telling.
sub do_tell {
    my $target = shift;
    my $msg = process_message (@_ ? "@_" : $_);

    my $cmd;
    if (defined $target) {
	my $num = to_num $target;
	if (defined $num && $num < 0) {
	    $cmd = 'it';
	}
	else {
	    $cmd = "tell $target";
	}
    }
    else {
	$cmd = 'shout';
    }

    if (defined $tell_queue_time) {
	# Make sure our tell won't be queued.
	my $time = time;
	my $queued_until = $last_tell_time + $tell_queue_time;
	sleep $queued_until - $time if $queued_until > $time;
    }

    my $ret = cmd "$cmd $msg";
    reset_tell;
    return $ret;
}

# Tell the user specified by the first argument the message
# specified by the second argument (or $_ if not given). Returns
# true on success or false on failure.
sub tell_raw {
    my $target = shift;
    my $msg = @_ ? "@_" : $_;

    $msg =~ s/\n+$//;

    return 1 if $msg =~ /^\s*$/;

    # Split it into several lines if needed.
    my @msg = ($msg);
    while (1) {
	my $str = $msg[$#msg];
	my $len = length $str;
	last if $len <= $max_tell;
	my $i;
	for ($i = $max_tell; $i >= 0; $i--) {
	    last if (substr $str, $i, 1) eq ' ';
	}
	$i = $max_tell if $i < 0;
	my $start = substr $str, 0, $i;
	my $end = substr $str, $i + 1;
	$msg[$#msg] = $start;
	push @msg, $end;
    }

    # Perform the tell.
    my $ret = 1;
    foreach (@msg) {
	return 0 unless do_tell $target;
    }
    return 1;
}

# Call tell_raw after sanitizing the message.
sub tell_now {
    my $target = shift;
    my $msg = @_ ? "@_" : $_;

    # Sanitize the message.
    $msg =~ s/\s+/ /g;
    $msg =~ s/^\s+//;
    $msg =~ s/\s+$//;

    tell_raw $target, $msg;
}

# Queue a tell_now for the current user.
sub tell {
    queue_for_user $current_user, \&tell_now, @_;
}

# Shout a message.
sub shout_now {
    tell_now undef, @_;
}
sub shout {
    tell undef, @_;
}

# Shout a message using the 'it' command.
sub it_now {
    tell_now -1, @_;
}
sub it {
    tell -1, @_;
}

# Process arguments for a qtell.
sub process_qtell {
    push @_, $_ unless @_;

    for (my $i = 0; $i < @_;) {
	my $str = $_[$i];
	my @msg = split /\\n|\n/, $str;
	push @msg, '' while $str =~ s/(?:\\n|\n)$//;
	push @msg, '' unless @msg;
	splice @_, $i, 1, @msg;
	$i += @msg;
    }

    # Replace tabs with the appropriate number of spaces.
    my $tabstop = 8;
    foreach my $str (@_) {
	my $len = length $str;
	for (my $i = 0; $i < $len; $i++) {
	    if ((substr $str, $i, 1) eq "\t") {
		my $tabwidth = $tabstop - ($i % $tabstop);
		substr $str, $i, 1, (' ' x $tabwidth);
		$_ += $tabwidth - 1 foreach $i, $len;
	    }
	}
    }

    return @_;
}

# A helper function for qtell and maybe_qtell.
sub qtell_raw {
    my $target = shift;

    # There's no reason to show trailing whitespace.
    s/\s+$// foreach @_;
    my $msg = join '\n', @_;

    # Make sure that the qtell isn't too long.
    my @msg;
    while ((length $msg) > $max_qtell) {
	my $i;
	for ($i = $max_qtell; $i > 0; $i--) {
	    last if (substr $msg, $i, 2) eq '\n';
	}
	$i = $max_qtell unless $i;
	push @msg, (substr $msg, 0, $i);
	$msg = substr $msg, $i + 2;
    }
    push @msg, $msg;

    my $ret;
    foreach (@msg) {
	# Skip empty messages.
	next if /^\s*$/;
	$ret = cmd "qtell $target $_";
    }
    reset_tell;
    return $ret;
}

# Perform a qtell. Each argument may have multiple lines, and multiple
# arguments will be treated as multiple lines.
sub qtell {
    my $target = shift;
    my $prev_allow_filter = $allow_filter;
    $allow_filter = 0;
    my $ret;
    if (defined $target) {
	$ret = qtell_raw $target, process_message process_qtell @_;
    }
    else {
	$ret = shout_now @_;
    }
    $allow_filter = $prev_allow_filter;
    return $ret;
}

# If the message is more than one line, call qtell, otherwise tell. If
# telling to a channel, make it look more like a channel tell.
sub maybe_qtell {
    my $target = shift;
    return shout @_ unless defined $target;
    my @msg = process_qtell @_;
    return tell $target, @msg unless @msg > 1;
    @msg = process_message @msg;

    if ($target =~ /^\d+$/) {
	# It's a channel tell. Use a regular tell unless it's channel
	# 120.
	return tell $target, @msg unless $target == 120;

	# Make it look like a channel tell.
	my $prefix = "$name(TD)($target): ";
	my $indent = ' ' x (length $prefix);
	my $first = 1;
	foreach (@msg) {
	    my $str = $first ? $prefix : $indent;
	    $first = 0;
	    $_ = $str . $_;
	}
    }

    queue_for_user $current_user, \&qtell_raw, $target, @msg;
}

# Get the variables of the user specified by the first argument (or
# $_) and return them as a hash. Returns an empty list on error.
sub get_server_vars {
    my $user = @_ ? $_[0] : $_;

    my @output = cmd "vars $user";
    return unless @output;

    if ($output[0] =~ /^Variable settings of (.+):$/ &&
	user_eq $1, $user) {
	shift @output;
    }
    else {
	warn "Got invalid variables for $user: $output[0]";
	return;
    }

    my %vars;
    foreach (@output) {
	s/^\r//;
	if (/^(\S+?): (.*)$/) {
	    $vars{$1} = $2;
	    next;
	}

	my @vars = split;
	shift @vars if @vars && $vars[0] eq '';
	foreach (@vars) {
	    next unless /^(.+?)=(.*)$/;
	    $vars{$1} = $2;
	}
    }

    return %vars;
}

# Get the width of a user for use in qtells.
sub get_width {
    my $user = @_ ? $_[0] : $_;

    my %vars = get_server_vars $user;
    my $width;
    if (exists $vars{width}) {
	$width = $vars{width};
    }
    else {
	warn "Cannot get width for $user";
	$width = 80;
    }

    # -1 for the ':'.
    return $width ? $width - 1 : 0;
}

# Perform some actions before we quit.
sub before_quit {
    if (defined $tell_queue_time) {
	# Sleep a bit to allow the last tells to go through.
	sleep $tell_queue_time;
    }
    $telnet->close;
    if (defined $log_file) {
	close $log_file or warn "Cannot close $log_filename: $!";
    }
    if (defined $debug_file) {
	close $debug_file or warn "Cannot close $debug_filename: $!";
    }
}

# Cause the bot to quit. Arguments are as for 'exit'.
sub quit {
    before_quit;
    print "Exiting\n" if $verbose;
    exit @_;
}

# Cause the bot to restart.
sub restart {
    before_quit;
    print "Restarting (executing '@cmdline')\n" if $verbose;
    exec $0 @cmdline;
    die "Cannot execute '@cmdline': $!";
}

# Superusers which can run special commands available only to them.
my @superusers = ($operator);

# Check if a given user is allowed to use superuser commands.
sub is_superuser {
    my $user = @_ ? $_[0] : $_;
    foreach (@superusers) {
	return 1 if user_eq $user, $_;
    }
    return 0;
}

# Check if a given user is the operator.
sub is_operator {
    my $user = @_ ? $_[0] : $_;
    user_eq $user, $operator;
}

# Get the handle of a user and whether or not that user is
# registered. In scalar context, returns the name of the user or undef
# if there is no such user or it's ambiguous. In list context, returns
# the empty list if there is no such user. If it's ambiguous, returns
# a list with the first element being true and the rest of the
# elements being the possible users. If not enough characters of the
# name were specified, return undef in scalar context or -1 in list
# context. Finally, if there is only one match in list context,
# returns a list of three elements, the first of which being false,
# the second of which being the name of the user, and the third of
# which being true if the user is registered, false otherwise.
sub get_user {
    my $user = @_ ? $_[0] : $_;

    return wantarray ? -1 : undef if length $user < 2;

    my @lines = cmd "finger $user";

    my @users;
    my $num_users;
    my $ambiguous = 0;
    my $could_parse = 0;
    foreach (@lines) {
	return
	    if /^There is no player matching the name \Q$user\E.$/ ||
	    (/^'(.*)' is not a valid handle.$/ && $user =~ /^\Q$1\E/);
	if (/^Finger of ($handle_pat)(.*):$/) {
	    return wantarray ? (0, $1, $2 ne '(U)') : $1
		unless $ambiguous;
	    $could_parse = 0;
	    last;
	}
	elsif (/^Ambiguous name \Q$user\E:$/) {
	    if ($ambiguous) {
		$could_parse = 0;
		last;
	    }
	    $ambiguous = 1;
	}
	elsif (/^-- Matches: (\d+) player\(s\) --$/) {
	    unless ($ambiguous) {
		$could_parse = 0;
		last;
	    }
	    $num_users = $1;
	}
	elsif ($ambiguous) {
	    push @users, split;
	    $could_parse = 1;
	}
	else {
	    $could_parse = 0;
	    last;
	}
    }

    # If we're here, it must be ambiguous (or a parse failure).
    $could_parse = $ambiguous && @users > 1 && @users == $num_users
	if $could_parse;

    unless ($could_parse) {
	warn q/Cannot parse 'finger' output; / .
	    "assuming $user doesn't exist";
	return;
    }

    return wantarray ? (1, @users) : undef;
}

# Match the 'needle' in 'haystack'. If 'case' is true, match
# case-sensitively, otherwise match case insensitively. Returns all
# matches (possibly none).
sub match {
    my %args = @_;
    my $needle = $args{needle};
    my @haystack = @{$args{haystack}};
    my $match_case = exists $args{case} ? $args{case} : 0;

    my $len = length $needle;
    for (my $i = 0; $i < $len; $i++) {
	last unless @haystack;
	my $c = substr $needle, $i, 1;
	for (my $j = 0; $j < @haystack;) {
	    my $d = substr $haystack[$j], $i, 1;
	    unless ($match_case) {
		$c = fc $c;
		$d = fc $d;
	    }
	    if ($c eq $d) {
		# Check for an exact match.
		return ($haystack[$j]) if $i eq ($len - 1) &&
		    (length $haystack[$j]) eq $len;
		$j++;
	    }
	    else {
		splice @haystack, $j, 1;
	    }
	}
    }

    return sort @haystack;
}

my %commands;

# Get usage strings for the command specified by the first argument.
sub get_usage {
    my $cmd = $_[0];
    my $usage = $commands{$cmd}->[1];
    my @usage = ((ref $usage) eq 'ARRAY') ? @$usage : ($usage);
    my @usage_lines;
    my $first = 1;
    foreach (@usage) {
	my $start = $first ? 'Usage:' : '  or: ';
	$first = 0;
	push @usage_lines, "$start $cmd" . ($_ ? " $_" : '');
    }
    return @usage_lines;
}

# Errors that occur while parsing options for commands.
my @opt_errors;

# A wrapper for GetOptionsFromArray so that error messages will be
# printed to the user who ran the command.
sub getopts {
    my $target = shift;
    local $SIG{__WARN__} = sub {
	my ($msg) = @_;
	chomp $msg;
	push @opt_errors, $msg;
    };
    GetOptionsFromArray @_;
}

# Parse arguments for callbacks. Returns a list containing the user
# the tell came from, the channel the user used to send the tell (or
# undef for direct tells), and either the channel the tell came from
# or the user if it was a direct tell, and finally the name of the
# command and the arguments given to it.
sub parse_args {
    my ($user, $channel, @rest) = @_;
    my $target;
    if (defined $channel) {
	if ($channel eq 'shout' || $channel eq 'it') {
	    $target = undef;
	}
	else {
	    $target = $channel;
	}
    }
    else {
	$target = $user;
    }
    return ($user, $channel, $target, @rest);
}

# Return a random one of the given args.
sub random_choice {
    die 'No args given' unless @_;
    return $_[int rand @_];
}

# Get all commands organized into regular, superuser, and operator
# commands, and sort them.
sub get_commands {
    my @cmds;
    my @sucmds;
    my @opcmds;
    foreach (keys %commands) {
	my $restrict = $commands{$_}->[3];
	if ($restrict == 0) {
	    push @cmds, $_;
	}
	elsif ($restrict == 1) {
	    push @sucmds, $_;
	}
	elsif ($restrict == 2) {
	    push @opcmds, $_;
	}
	else {
	    die "Invalid restriction value: $restrict";
	}
    }
    my @ret = (\@cmds, \@sucmds, \@opcmds);
    @$_ = sort @$_ foreach @ret;
    return @ret;
}

# Return help for commands and variables. The first argument is the
# maximum length, the second argument is a subroutine to get help, and
# the rest are the things to get help for.
sub get_help {
    my ($maxlen, $get_help, @things) = @_;
    my @help;
    foreach my $thing (@things) {
    	my $help = $get_help->($thing);
    	$help = (split "\n", $help)[0];
    	$help =~ s/\.$//;
    	$help = lcfirst $help;
    	my $space = ' ' x ($maxlen - length $thing);
    	my $start = "  $thing$space   ";
    	my $tab = ' ' x ((length $start) + 2);
    	push @help, (wrap '', $tab, ($start . $help));
    }
    return @help;
}

sub get_commands_help {
    my ($maxlen, @cmds) = @_;
    get_help $maxlen, sub { $commands{$_[0]}->[2] }, @cmds;
}

sub get_vars_help {
    my ($maxlen, @vars) = @_;
    get_help $maxlen, sub { $variables{$_[0]}->{help} }, @vars;
}

# A helper function to check a user for the do_set function and the
# 'vars' command. Returns the name of the user on success, or undef on
# error.
sub check_var_user {
    my ($user, $target, $var_user) = @_;
    my ($ambiguous, @users) = get_user $var_user;
    unless (defined $ambiguous) {
	qtell $target, "No such user: $var_user";
	return undef;
    }
    if ($ambiguous < 0) {
	qtell $target, 'You need to specify at least two ' .
	    'characters of the name.';
	return undef;
    }
    if ($ambiguous) {
	local $Text::Wrap::columns = get_width $user;
	qtell $target, "Ambiguous name $var_user:",
	    '-- Matches: ' . (scalar @users) . ' players --',
	    (wrap '', '', @users);
	return undef;
    }
    my ($name, $registered) = @users;
    if (! $registered) {
	qtell $target, 'Sorry, but unregistered users ' .
	    'cannot have variables at this time.';
	return undef;
    }
    return $name;
}

# A helper function for the 'set' and 'setvar' commands.
sub do_set {
    my $other_user = shift;
    my ($user, $channel, $target, $me, @args) = parse_args @_;

    my $var_user = $other_user ? shift @args : $user;
    return 0 if @args > 2;
    my ($key, $value) = @args;
    return 0 unless (defined $var_user) && (defined $key);
    $value = '' unless defined $value;

    return 1 unless defined ($var_user = check_var_user
			     $user, $target, $var_user);

    unless (exists $variables{$key}) {
	qtell $target, "Invalid variable: '$key'";
	return 1;
    }

    my %spec = %{$variables{$key}};
    my $type = $spec{type};
    my (@values, @range);
    foreach ([values => \@values], [range => \@range]) {
	my ($name, $var) = @$_;
	@$var = @{$spec{$name}} if exists $spec{$name};
    }

    if ($type eq 'bool') {
	if ($value =~ /^(?:on|yes|true)$/i) {
	    $value = 1;
	}
	elsif ($value =~ /^(?:off|no|false)$/i) {
	    $value = 0;
	}
    }

    unless (var_is_valid $key, $value) {
	if ($value ne '') {
	    my @msg = ("Invalid value for $key: $value");
	    if (@range) {
		my ($lower, $upper) = @range;
		my $type_name;
		if ($type eq 'int') {
		    $type_name = 'an integer'
		}
		elsif ($type eq 'float') {
		    $type_name = 'floating point number';
		}
		else {
		    die "Invalid numeric type: $type";
		}
		push @msg, "Must be $type_name between $lower and " .
		    "$upper inclusive";
	    }
	    if (@values) {
		my @v = @values;
		@v = map { "'$_'" } @v if $type eq 'string';
		my $values = join ', ', @v;
		push @msg, "Possible values: $values";
	    }
	    local $Text::Wrap::columns = get_width $user;
	    $_ = wrap '', '', $_ foreach @msg;
	    qtell $target, @msg;
	}
	else {
	    qtell $target, "The $key variable may not be empty";
	}
	return 1;
    }

    $value += 0 if is_numeric_type $type;

    if ($other_user) {
	if ($value ne '') {
	    qtell $target, "Set $var_user\'s $key variable to $value";
	    tell_now $var_user,
		"$user set your $key variable to $value";
	}
	else {
	    qtell $target, "Unset $var_user\'s $key variable";
	    tell_now $var_user,
		"$user unset your $key variable";
	}
    }
    else {
	if ($value ne '') {
	    qtell $target, "Set $key to $value";
	}
	else {
	    qtell $target, "Unset $key";
	}
    }
    set_vars $var_user, $key => $value or qtell $target,
	"Error saving variables; please contact $operator";

    return 1;
}

# Used by cowsay and cowthink.
my $cowsay_usage =
    '[-bdgpstwy] [-e EYES] [-T TONGUE] [-f COWFILE] [-r] MESSAGE';
my $cowsay_help = <<EOF;
 like a cow.

  -b           initiate borg mode
  -d           cause the cow to appear dead
  -g           invoke greedy mode
  -p           cause a state of paranoia to come over the cow
  -s           make the cow appear thoroughly stoned
  -t           yield a tired cow
  -w           somewhat the opposite of -t, and initiates wired mode
  -y           brings on the cow's youthful appearance
  -e EYES      set the cow's eyes to EYES
  -T TONGUE    set the cow's tongue to TONGUE
  -f COWFILE   use COWFILE as the cow
  -r           use a random cowfile

Available cowfiles: @cowsay_cowfiles.

The homepage for cowsay is no longer available, but you can find it on the Web Archive here: https://web.archive.org/web/20120527202447/http://www.nog.net/~tony/warez/cowsay.shtml
EOF

sub cowsay {
    my $cowsay_prog = 'cow' . shift;
    my ($user, $channel, $target, $me, @args) = parse_args @_;
    my $mode;
    my $cowfile;
    return 0 unless getopts $target, \@args,
	'b'	=> sub { $mode = 'b' },
	'd'	=> sub { $mode = 'd' },
	'g'	=> sub { $mode = 'g' },
	'p'	=> sub { $mode = 'p' },
	's'	=> sub { $mode = 's' },
	't'	=> sub { $mode = 't' },
	'w'	=> sub { $mode = 'w' },
	'y'	=> sub { $mode = 'y' },
	'e=s'	=> \my $eyes,
	'T=s'	=> \my $tongue,
	'f=s'	=> \$cowfile,
	'r'	=> sub { $cowfile = '' };
    return 0 unless @args;

    if (defined $cowfile) {
	if ($cowfile eq '') {
	    $cowfile = random_choice @cowsay_cowfiles;
	}
	elsif ($cowfile =~ m|/|) {
	    tell $target, 'Cowfile must not be a full path';
	    return 1;
	}
    }

    # -3 for the speech/thought bubble boundaries.
    my $width = (get_width $user) - 3;
    $width = 0 if $width < 0;

    my @cmd = ($cowsay_prog, '-W', $width);
    push @cmd, "-$mode" if defined $mode;
    push @cmd, '-e', $eyes if defined $eyes;
    push @cmd, '-T', $tongue if defined $tongue;
    push @cmd, '-f', $cowfile if defined $cowfile;
    push @cmd, '--', @args;

    my $cowsay;
    unless (open $cowsay, '-|', @cmd) {
	tell $target, "Cannot run the $cowsay_prog command. " .
	    "Please contact $operator.";
	return 1;
    }
    my @message = <$cowsay>;
    unless (close $cowsay) {
	tell $target, "Invalid cowfile: '$cowfile'";
	return 1;
    }

    chomp foreach @message;

    qtell $target, @message;
    return 1;
}

# Print the help for a command or variable. Returns true if a single
# command/variable was found, or false otherwise.
sub do_help {
    my ($user, $target, $thing, $type, $get_usage,
	$get_help, $get_restricted, @things) = @_;
    my $Type = ucfirst $type;

    my @matches = match
	needle		=> $thing,
	haystack	=> \@things;

    unless (@matches) {
	qtell $target, "No help for $type $thing.";
	return 0;
    }
    elsif (@matches > 1) {
	local $Text::Wrap::columns = get_width $user;
	qtell $target, (wrap '', '', "$Type '$thing' is ambiguous; " .
			"possibilities: @matches");
	return 0;
    }

    $thing = $matches[0];
    my $help = $get_help->($thing);
    my $restricted = $get_restricted->($thing);
    chomp $help;
    local $Text::Wrap::columns = get_width $user;
    $help = wrap '', '', $help;

    my @help;
    push @help, 'RESTRICTED TO ' .
	(($restricted == 1) ? 'SUPERUSERS' : 'OPERATORS')
	if $restricted;
    push @help, $get_usage->($thing);
    push @help, $help;
    qtell $target, @help;

    return 1;
}

# The commands available to users. Each value of the %commands hash is
# an array reference, the first item of which is the command callback,
# the second item of which is a short string describing the usage of
# that command, and the third item of which is a longer help
# string. The fourth item of the array reference is 0 if anyone can
# run the command, 1 if superuser permission is needed, or 2 if only
# the operator can run the command. The callbacks are called with the
# first argument as the user that ran the command, the second argument
# as the name the command was called as, and the rest of the arguments
# as the arguments given to the command. The callbacks should return
# true on success, or false for invalid syntax.
%commands = (
    quit => [
	sub {
	    my ($user, $channel, $target, $me, @args) = parse_args @_;
	    return 0 if @args > 0;
	    qtell $target, 'Shutting down...';
	    tell_now $operator, "$user shut me down.";
	    quit;
	},
	'', "Make $name exit the server.", 1
    ],
    restart => [
	sub {
	    my ($user, $channel, $target, $me, @args) = parse_args @_;
	    return 0 if @args > 0;
	    qtell $target, 'Restarting...';
	    tell_now $operator, "$user restarted me.";
	    $ENV{$restart_target_var} =
		(defined $target) ? $target : '';
	    restart;
	}, '', "Restart $name.", 1
    ],
    setvar => [
	sub {
	    do_set 1, @_;
	},
	'USER VARIABLE [VALUE]', q/Set USER's VARIABLE to VALUE./, 1
    ],
    run => [
	sub {
	    my ($user, $channel, $target, $me, @args) = parse_args @_;
	    return 0 unless @args;
	    my @out = cmd @args;
	    push @out, "Cannot run '@args'; please contact $operator."
		unless @out;
	    qtell $target, @out;
	    return 1;
	}, 'COMMAND [ARG]...', 'Run the given server COMMAND.', 1
    ],
    queue => [
	sub {
	    my ($user, $channel, $target, $me, @args) = parse_args @_;
	    my $mode;
	    my $when;
	    return 0 unless getopts $target, \@args,
		'd|delay=f'	=> sub {
		    $mode = 'relative';
		    $when = $_[1];
		},
		'w|when=s'	=> sub {
		    $mode = 'absolute';
		    # It is not possible to pass spaces in arguments,
		    # so allow underscores instead.
		    $when = $_[1] =~ s/_/ /gr;
		};
	    return 0 unless @args;

	    my $now = time;

	    if (defined $when) {
		if ($mode eq 'relative') {
		    if ($when < 0) {
			qtell $target, "DELAY cannot be negative.";
			return 1;
		    }
		}
		else {
		    my ($time, $error) = parsedate $when,
			NOW => $now, WHOLE => 1;
		    if (defined $time && $time < $now) {
			$error = (strftime '%c', localtime $time) .
			    ' is in the past';
			undef $time;
		    }
		    unless (defined $time) {
			local $Text::Wrap::columns = get_width $user;
			qtell $target,
			    (wrap '', '',
			     "Invalid WHEN '$when': $error.");
			return 1;
		    }
		    $when = $time;
		}
	    }
	    else {
		$mode = 'relative';
		($when) = get_vars $user, 'delay';
	    }

	    my $msg = "Queued '@args' to run ";
	    if ($mode eq 'relative') {
		local $Time::Duration::MILLISECOND = 1;
		$msg .= 'in ' . (duration $when);
		$when += $now;
	    }
	    else {
		$msg .= 'at ' . (strftime '%c', localtime $when);
	    }
	    $msg .= '.';

	    queue_absolute $when, $user, 0, \&cmd, @args;

	    local $Text::Wrap::columns = get_width $user;
	    qtell $target, (wrap '', '', $msg);

	    return 1;
	},  '[-d DELAY|-w WHEN] COMMAND [ARG]...', <<EOF, 1
Queue COMMAND to run WHEN or after DELAY seconds.

If neither DELAY nor WHEN is given, use the value of your 'delay' variable. WHEN, if given, is the absolute time at which to run the command.

  -d, --delay=DELAY   the delay until COMMAND is run
  -w, --when=WHEN     when COMMAND should be run
EOF
    ],
    suadd => [
	sub {
	    my ($user, $channel, $target, $me, @args) = parse_args @_;
	    return 0 if @args < 1;
	    foreach (@args) {
		if (is_superuser) {
		    qtell $target, "$_ is already a superuser.";
		}
		else {
		    push @superusers, $_;
		    qtell $target, "$_ added to the superuser list.";
		    tell $_, 'You have been added to the ' .
			"superuser list by $user. Please don't " .
			'abuse your new powers.';
		}
	    }
	    return 1;
	}, 'USER...', 'Temporarily add USERs to the superuser list.', 2
    ],
    sudel => [
	sub {
	    my ($user, $channel, $target, $me, @args) = parse_args @_;
	    return 0 if @args < 1;
	    foreach (@args) {
		my $i;
		for ($i = 0; $i < @superusers; $i++) {
		    last if user_eq $user, $superusers[$i];
		}
		if ($i >= @superusers) {
		    qtell $target, "$_ is not a superuser.";
		}
		elsif (is_operator $user) {
		    qtell $target, 'Cannot remove the operator ' .
			(@args > 1 ? "($operator) " : '') .
			'from the superuser list.';
		}
		else {
		    splice @superusers, $i, 1;
		    qtell $target, "$_ removed from the superuser list.";
		    tell $_, 'You have been removed from the ' .
			"superuser list by $user.";
		}
	    }
	    return 1;
	}, 'USER...',
	'Temporarily remove USERs from the superuser list.', 2
    ],
    set => [
	sub {
	    do_set 0, @_;
	}, 'VARIABLE [VALUE]', 'Set VARIABLE to VALUE.', 0
    ],
    vars => [
	sub {
	    my ($user, $channel, $target, $me, @args) = parse_args @_;
	    my $vars_user = @args ? shift @args : $user;
	    return 0 if @args;

	    return 1 unless defined ($vars_user = check_var_user
				     $user, $target, $vars_user);

	    my %vars = get_vars $vars_user;

	    my @vars = map {
		my $type = $variables{$_}->{type};
		my $value = $vars{$_};
		$value = "'$value'" if $type eq 'string';
		["  $_=$value   ", $type];
	    } @variables;

	    my $maxlen;
	    foreach (@vars) {
		my $len = length $_->[0];
		$maxlen = $len if ! defined $maxlen || $len > $maxlen;
	    }

	    qtell $target, "Variable settings of $vars_user:", map {
		my ($str, $type) = @$_;
		$str . (' ' x ($maxlen - length $str)) . "($type)";
	    } @vars;

	    return 1;
	}, '[USER]', 'Print the variables for USER or yourself if ' .
	'USER is omitted.', 0
    ],
    flush => [
	sub {
	    my ($user, $channel, $target, $me, @args) = parse_args @_;
	    return 0 unless getopts $target, \@args,
		'a|all' => \my $all;
	    my $is_sup = is_superuser $user;
	    if (! $is_sup && $all) {
		qtell $target,
		    'Only superusers may use the -a option.';
		return 1;
	    }
	    my $flush_user = $user;
	    if (@args) {
		unless ($is_sup) {
		    qtell $target, 'Only superusers can flush other ' .
			q/users' communications./;
		    return 1;
		}
		return 0 if $all;
		$flush_user = shift @args;
		return 0 if @args;
	    }

	    if ($all) {
		flush_queue;
	    }
	    else {
		flush_queue $flush_user;
	    }

	    return 1;
	}, ['[USER]', '-a'], <<EOF, 0
Flush pending communications.

USER, if given, is the user for which to flush communications. If not given, defaults to you. Only superusers may specify USER or use the -a option.

  -a, --all   flush all communications for all users
EOF
    ],
    figlet => [
	sub {
	    my ($user, $channel, $target, $me, @args) = parse_args @_;
	    my $font = $figlet_default_font;
	    my @opts = ('f=s' => \$font);
	    my $justification;
	    foreach my $opt (qw(c l r x)) {
		push @opts, $opt => sub { $justification = $opt };
	    }
	    push @opts, 'w=i' => \my $width;
	    my $kerning;
	    foreach my $opt (qw(s S k W o)) {
		push @opts, $opt => sub { $kerning = $opt };
	    }
	    my $direction;
	    foreach my $opt (qw(L R X)) {
		push @opts, $opt => sub { $direction = $opt };
	    }
	    return 0 unless (getopts $target, \@args, @opts) && @args;

	    if ($font =~ m|/|) {
		qtell $target, 'Font must not be a full path';
		return 1;
	    }

	    if (defined $width) {
		if ($width <= 0) {
		    qtell $target, 'Width must be positive';
		    return 1;
		}
	    }
	    else {
		$width = get_width $user;
	    }

	    my @cmd = ('figlet', '-w', $width, '-f', $font);
	    foreach ($justification, $kerning, $direction) {
		push @cmd, "-$_" if defined;
	    }
	    push @cmd, '--', @args;

	    my $figlet;
	    unless (open $figlet, '-|', @cmd) {
		qtell $target, 'Cannot run the figlet command. ' .
		    "Please contact $operator.";
		return 1;
	    }
	    my @message = <$figlet>;
	    unless (close $figlet) {
		qtell $target, "Invalid font: '$font'";
		return 1;
	    }

	    foreach (@message) {
		chomp;
		if (/[^[:print:]]/) {
		    qtell $target,
			"The font you chose ($font) cannot be " .
			'transferred. Please try a different font.';
		    return 1;
		}
	    }

	    qtell $target, @message;
	    return 1;
	}, '[-clrxsSkWoLRX] [-f FONT] [-w WIDTH] MESSAGE', <<EOF, 0
Print a message in ASCII art.

  -f FONT    print the message in font FONT (see available fonts)
  -c         center the output horizontally
  -l         make the output flush-left
  -r         make the output flush-right
  -x         set justification according to the text direction
  -w WIDTH   set the output width to WIDTH (default, 'width' variable)
  -s         smush characters as close as possible
  -S         smush characters even closer than possible
  -k         kern, i.e., make characters close without smushing
  -W         display characters at their full width
  -o         overlap characters
  -L         print text left-to-right
  -R         print text right-to-left
  -X         use whatever direction is the default for the font

Available fonts: @figlet_fonts.

The figlet homepage is here: http://www.figlet.org/
EOF
    ],
    cowsay => [
	sub {
	    return cowsay 'say', @_;
	}, $cowsay_usage, 'Talk' . $cowsay_help, 0
    ],
    cowthink => [
	sub {
	    return cowsay 'think', @_;
	}, $cowsay_usage, 'Think' . $cowsay_help, 0
    ],
    filter => [
	sub {
	    my ($user, $channel, $target, $me, $filter, @args) =
		parse_args @_;
	    return 0 unless (defined $filter) && @args;

	    my @matches = match
		needle		=> $filter,
		haystack	=> \@filter_progs;
	    if (@matches > 1) {
		local $Text::Wrap::columns = get_width $user;
		qtell $target,
		    (wrap '', '', "Filter '$filter' is " .
		     "ambiguous; possibilities: @matches");
		return 1;
	    }
	    elsif (@matches) {
		$filter = $matches[0];
	    }
	    # If there are no matches, let do_filter handle it.

	    my $output = do_filter $filter,
		sub { qtell $target, "@_" }, "@args";
	    qtell $target, $output if defined $output;
	    return 1;
	}, 'FILTER MESSAGE', <<EOF, 0
Filter MESSAGE through FILTER and print the output.

Available filters:

$filters_help

All of the filters except for pig come from the filters package by Joey Hess. The homepage is here: https://joeyh.name/code/filters/ The pig filter comes from bsdgames. There doesn't seem to be an official homepage for bsdgames, but a web search is sure to turn up something.
EOF
    ],
    etime => [
	sub {
	    my ($user, $channel, $target, $me, @args) = parse_args @_;
	    return 0 if @args < 1 || @args > 2;
	    my ($time, $inc) = @args;

	    # Verify the arguments.
	    foreach ([\$time, 'TIME'], [\$inc, 'INC']) {
		my ($var, $name) = @$_;
		unless (defined $$var) {
		    $$var = 0;
		    next;
		}
		my $value = to_num $$var;
		unless (defined $value && $value == int $value) {
		    qtell $target, "$name is not an integer: $$var";
		    return 1;
		}
		if ($value < 0) {
		    qtell $target,
			"$name must not be negative: $value";
		    return 1;
		}
		$$var = $value;
	    }

	    # If $time is 0, make it 10 seconds. Otherwise, convert it
	    # to seconds.
	    $time = $time ? $time * 60 : 10;

	    # Calculate the etime. The 'TIME + INC*2/3' formula actual
	    # mixes units, so this is the actual formula.
	    my $etime = $time + ($inc * 40);

	    # Get the category. Remember, $etime is in seconds, so
	    # Lightning is less than 3 minutes == 180 seconds, etc.
	    my $category;
	    if ($etime < 180) {
		$category = 'Lightning';
	    }
	    elsif ($etime < 900) {
		$category = 'Blitz';
	    }
	    else {
		$category = 'Standard';
	    }

	    # Print the result.
	    qtell $target, ("Expected time:\t" . duration $etime),
		"Category:\t$category";

	    return 1;
	}, 'TIME [INC]', <<EOF, 0
Calculate the expected time for a game.

TIME is given in minutes and INC is given in seconds, and they must be whole numbers. A TIME of 0 actually means 10 seconds. If INC is omitted, it will default to 0.

The formula for calculating etime is: TIME + INC*2/3.

See also 'help etime'.
EOF
    ],
    superusers => [
	sub {
	    my ($user, $channel, $target, $me, @args) = parse_args @_;
	    return 0 if @args;
	    qtell $target, "Superusers: @superusers";
	    return 1;
	}, '', 'List all superusers.', 0
    ],
    help => [
	sub {
	    my ($user, $channel, $target, $me, @args) = parse_args @_;
	    my ($reg, $sup, $op, $vars);
	    return 0 unless getopts $target, \@args,
		'r|regular'		=> \$reg,
		's|superuser'		=> \$sup,
		'o|operator'		=> \$op,
		'v|vars|variables'	=> \$vars,
		'a|all'			=>
		sub { $reg = $sup = $op = $vars = 1 };
	    return 0 if @args > 1 ||
		(@args && ($reg || $sup || $op || $vars));

	    if (@args) {
		my $item = $args[0];
		if ($item =~ /^v_(.*)$/i) {
		    if ($1 eq '') {
			qtell $target,
			    'Variable name cannot be empty.';
			return 1;
		    }
		    do_help $user, $target, $1, 'variable', sub {
			my ($var) = @_;
			my %spec = %{$variables{$var}};
			my $type = $spec{type};
			my $help = "Variable: $var ($type)";
			if (exists $spec{range}) {
			    my ($lower, $upper) = @{$spec{range}};
			    $help .= " [$lower .. $upper]";
			}
			if (exists $spec{default}) {
			    my $default = $spec{default};
			    $default = "'$default'"
				if $type eq 'string';
			    $help .= " (default: $default)"
			}
			return $help;
		    }, sub { $variables{$_[0]}->{help} }, sub { 0 },
			@variables;
		}
		else {
		    do_help $user, $target, $item, 'command',
			\&get_usage, sub { $commands{$_[0]}->[2] },
			sub { $commands{$_[0]}->[3] }, keys %commands;
		}
		return 1;
	    }

	    # It would be cleaner to do
	    #     \my (@cmds, @sucmds, @opcmds) = get_commands;
	    # But that feature is experimental, so we'd better not...
	    my ($cmds, $sucmds, $opcmds) = get_commands;
	    my @cmds = @$cmds;
	    my @sucmds = @$sucmds;
	    my @opcmds = @$opcmds;

	    unless ($reg || $sup || $op || $vars) {
		$reg = 1;
		$sup = is_superuser $user;
		$op = is_operator $user;
		$vars = 1;
	    }

	    my $maxlen;
	    my @things;
	    push @things, @cmds if $reg;
	    push @things, @sucmds if $sup;
	    push @things, @opcmds if $op;
	    push @things, @variables if $vars;
	    foreach (@things) {
		my $len = length;
		$maxlen = $len if ! defined $maxlen || $len > $maxlen;
	    }

	    local $Text::Wrap::columns = get_width $user;
	    my @help;
	    push @help, 'The following commands can be run through ' .
		'direct tells or through tells to channel 120. ' .
		'Commands may be abbreviated as long as they are ' .
		'unambiguous.' if $reg || $sup || $op;
	    push @help, q/Variables are set using the 'set' command./
		if $vars;
	    @help = wrap '', '', (join ' ', @help);
	    push @help, "\nCommands:",
		(get_commands_help $maxlen, @cmds) if $reg;
	    push @help, "\nSuperuser commands:",
		(get_commands_help $maxlen, @sucmds) if $sup;
	    push @help, "\nOperator commands:",
		(get_commands_help $maxlen, @opcmds) if $op;
	    push @help, "\nVariables:",
		(get_vars_help $maxlen, @variables) if $vars;
	    my @help_cmds;
	    push @help_cmds, '<COMMAND>' if $reg || $sup || $op;
	    push @help_cmds, 'v_<VARIABLE>' if $vars;
	    $_ = "'help $_'" foreach @help_cmds;
	    my $help_cmds = join ' or ', @help_cmds;
	    push @help,
		(wrap '', '', "\nTry $help_cmds for more info.");
	    qtell $target, @help;

	    return 1;
	},
	['[-rsova]', '[COMMAND|v_VARIABLE]'], <<EOF, 0
Print the help for COMMAND or VARIABLE.

If no COMMAND or VARIABLE is given, print a summary of some or all commands and variables. No options may be given if COMMAND or VARIABLE is given.

  -r, --regular     include regular commands
  -s, --superuser   include superuser commands
  -o, --operator    include operator commands
  -v, --variables   include variables
  -a, --all         include everything (equivalent to -rsov)

If no options are given, include all variables and all commands that you have permission to run.
EOF
    ]
);

# Get a message that blik used to say.
sub blik_phrase {
    my $fortune;
    unless (open $fortune, '-|', 'fortune', '-n', $max_fortune, '-s',
	    '--', $blik_phrase_file) {
	warn "Cannot run 'fortune': $!";
	return undef;
    }
    my $phrase = <$fortune>;
    unless (close $fortune) {
	my $signal = $? & 127;
	if ($signal) {
	    my $coredump = $? & 128;
	    warn "'fortune' killed by signal $signal" .
		($coredump ? ' with a coredump' : '');
	}
	my $status = $? >> 8;
	if ($status) {
	    warn "'fortune' exited with status $status";
	}
	else {
	    warn q/An unknown error occured when running 'fortune'./;
	}
	return undef;
    }
    chomp $phrase;
    return $phrase;
}

my $good = qr/good|great|awsome|magnificent|cool/;

# These are like commands, except that they match a pattern rather
# than a specific string. The string to match against is always
# stripped of punctuation, and spaces are normalized (i.e.,
# s/\s+/ /). @phrases is an array rather than a hash to preserve the
# order of the keys, since the order in which patterns are tried
# matters. Each key is a pattern and each value is a callback. In the
# callbacks, $1, $2, etc. are available as groups in the pattern. If
# the callbacks return false, it will be as though they were never
# matched. Otherwise, they should return true.
my @phrases = (
    qr/\b(?:what(?:s| is) $name|(?:what do you do|what is your function))\b/i => sub {
	my ($user, $channel, $target) = @_;
	tell $target, 'I print random phrases, and sometimes even ' .
	    'relevant ones, like this one!';
    },
    qr/\b(?:(hit|poke|kick|kill|boot|slap|elbow|knee|head[- ]?(?:butt|bump))s?|(punch)(?:es)?) (?:$name|you)\b/i => sub {
	my ($user, $channel, $target, @m) = @_;

	my $action;
	my $s;
	if (defined $m[0]) {
	    $action = $m[0];
	    $s = 's';
	}
	elsif (defined $m[1]) {
	    $action = $m[1];
	    $s = 'es';
	}
	else {
	    tell $target, 'I think you broke me! Please tell ' .
		"$operator so he can fix me. No seriously, I'm " .
		'actually broken. This should not be happening.';
	    warn 'Unknown action';
	    return;
	}
	my $ing = ($action =~ /^[^t]t$/i) ? 'ting' : 'ing';

	my @it_actions = ('right back', 'in the face');
	my $answer = random_choice 'Ouch!', 'That hurts!',
	    'Stop that!', q/That's not funny!/,
	    "Stop $action$ing me!", ((undef) x @it_actions);
	if (defined $answer) {
	    tell $target, $answer;
	    return;
	}

	$answer = ((defined $channel) ? $user : 'you') . ' ' .
	    (random_choice @it_actions) . '!';
	if (defined $target) {
	    tell $target, "I $action $answer";
	}
	else {
	    it "$action$s $answer";
	}
    },
    qr/\b(?:marry me|(?:wanna|want to) get married)\b/i => sub {
	my ($user, $channel, $target) = @_;
	tell $target,
	    (random_choice q/Why of course! I thought you'd never / .
	     "ask! ...Oh wait, you're $user? Never mind I thought " .
	     'you were Lect--er, someone else...',
	     'Ha! You must be joking.',
	     "Marry YOU, $user? I'd rather die!",
	     "No, but I think $operator might! (Hee hee hee, this " .
	     'ought to be pretty funny! Wait, did I say that out ' .
	     'loud?)');
    },
    qr/\b(?:what do you think of blik|asks $name what (?:it|s?he) thinks of blik|(?:does|did|do) (?:$name|you)(?: ever)? know blik)\b/i => sub {
	my ($user, $channel, $target) = @_;
	tell $target, 'Although I never met blik personally, blik ' .
	    'was my inspiration. I will always look up to blik.';
    },
    qr/\b(?:$name is not(?: .+)? as $good as blik(?: (?:was|is)|s)?|blik(?: (?:was|is)|s)?(?: .+) better than $name|(?:you are|your) not(?: .+)? blik)\b/i => sub {
	my ($user, $channel, $target) = @_;
	tell $target,
	    (random_choice "Hopefully you find me sufficient for " .
	     "now, $user.", "I'm always improving!");
    },
    qr/\b(?:$name is better than blik|(?:are you|you are) better than blik)\b/i => sub {
	my ($user, $channel, $target) = @_;
	tell $target,
	    (random_choice
	     q/Well I'm pretty good, but then so was blik.../,
	     q/Some of you don't seem to think so./,
	     q/Well I haven't actually met blik, so I don't know./);
    },
    qr/\bshould not be allowed|do not think (?:$name|you) should be allowed\b/i => sub {
	my ($user, $channel, $target) = @_;
	tell $target,
	    (random_choice
	     q/Yes, that's probably a good idea. *sigh*/,
	     q/I don't think I could manage that anyway./,
	     'Yes, things like that should probably be left to humans.');
    },
    qr/\b(?:$name|you) should be allowed\b/i => sub {
	my ($user, $channel, $target) = @_;
	tell $target,
	    (random_choice
	     q/You think so? I'm not sure that would be a good idea.../,
	     q/I don't know if I have the skills for that./,
	     q/That sounds like a human job./);
    },
    qr/\b(?:are you bored|how bored are you)\b/i => sub {
	my ($user, $channel, $target) = @_;
	my @msgs = ("Not too bored, $user.", @bored_messages);
	my $msg = bored_message $boredom, @msgs;
	$msg .= ' Thanks for' . (($boredom > 1) ? ' finally' : '') .
	    " saying something, $user" .
	    (($boredom > 2) ? ('!' x ($boredom - 2)) : '.')
	    if $boredom;
	tell $target, $msg;
    },
    qr/\bloves? (?:$name|you)\b/i => sub {
	my ($user, $channel, $target) = @_;
	tell $target,
	    (random_choice "Aww, I love you too, $user. <3",
	     q/You're too kind./, 'How sweet. :-}');
    },
    qr/\bdo(?:es)? (?:$name|you) (dis)?agree\b/i => sub {
	my ($user, $channel, $target, @m) = @_;
	tell $target,
	    (random_choice
	     ((defined $m[0]) ?
	      ("Why of course, $user! What a ridiculous thing to say!",
	       q/Hmm... I'm not too sure.../, 'Well, I dunno...',
	       'Certainly not! I agree with that 100%!') :
	      ('Certainly!', 'Nah, not really.',
	       "Most certainly not, $user! How ridiculous!")));
    },
    qr/\b(?:$name|you) (?:(?:are|is) boring|bores? me)\b/i => sub {
	my ($user, $channel, $target) = @_;
	tell $target,
	    (random_choice "Well I think YOU'RE boring, $user.",
	     q/Do I bore you? I didn't realize./,
	     "I'm sorry you think I'm boring, $user.");
    },
    qr/\bsquirrel(s)? in(?:side)?(?: of)? (?:$name|you)\b/i => sub {
	my ($user, $channel, $target, @m) = @_;

	my $squirrel;
	my $it;
	if (defined $m[0]) {
	    $squirrel = 'those squirrels';
	    $it = 'them';
	}
	else {
	    $squirrel = 'that squirrel';
	    $it = 'it';
	}
	my $you = (defined $channel) ? $user : 'you';

	my $answer = random_choice
	    'What is it with stuffing squirrels inside me?',
	    'The squirrel is a noble beast. For on acorns does it ' .
	    'feast. And it, um, runs up trees... Maybe I ' .
	    q/shouldn't be a poet.../, undef;

	if (defined $answer) {
	    tell $target, $answer;
	}
	elsif (defined $target) {
	    tell $target, "Well I take $squirrel out and put $it" .
		" right inside $you!";
	}
	else {
	    it "takes $squirrel out and puts $it right inside $you!";
	}

    },
    qr/(?:\bgives? (?:$name|you)|^have)(?: (?:some(?: more)?(?: of the)?|(?:.+ )?more|the|an?|another))? (.+)$/i => sub {
	my ($user, $channel, $target, @m) = @_;
	tell $target,
	    (random_choice "Thanks for the $m[0], $user!",
	     "Thanks, $user! I will always treasure the $m[0].");
    },
    sub { /\bwant to play (chess hangman|connect four)\b/i } => sub {
	my ($user, $channel, $target, @m) = @_;

	unless (user_eq $user, 'LectureNova') {
	    # An impostor!
	    tell $target,
		(random_choice
		 "What, are you LectureNova now, $user?",
		 'We already have one LectureNova, and that is ' .
		 'more than enough.');
	    return;
	}

	unless (defined $channel && $channel eq 'shout') {
	    tell $target,
		"What?!? I thought you only shout that, $user!";
	    return;
	}

	# Only respond 1/2 of the time so we're not too annoying.
	return unless (rand 2) > 1;

	my $cmd = random_choice qw(figlet cowsay cowthink filter);
	$cmd .= ' ' . random_choice keys %filter_progs
	    if $cmd eq 'filter';
	my $cmds = qq/"+ch 120" and "tell 120 $cmd <MESSAGE>"/;
	tell $target,
	    (random_choice
	     "Forget about $user! Try $cmds instead!",
	     "That's nice, $user, but if you want to have REAL " .
	     "fun, you should $cmds.",
	     ((ucfirst $m[0]) . ' is all fine and good, but you ' .
	      "should try $cmds."));
    },
    qr/(?:^chess(?: $name)?$)|\bplays?(?: .+)? chess\b/i => sub {
	my ($user, $channel, $target) = @_;
	tell $target,
	    (random_choice
	     q/Actually, I don't know how to play chess.../,
	     'I think FICS has enough computer players already.',
	     q/My programmer isn't competent enough to write a / .
	     'decent chess engine. Or a decent chat bot, for that ' .
	     'matter.');
    },
    qr/\b(fail(?:ed|s)?|pass(?:e[ds])?)(?: the)? turings? tests?\b/i => sub {
	my ($user, $channel, $target, @m) = @_;
	my @choices = (q/Actually, I'm secretly a human./);
	if ($m[0] =~ /^fail/i) {
	    push @choices, q/I would only fail because I'm so much / .
		"smarter than any human, $user (especially you).",
		'Well I think YOU might fail the Turing Test!';
	}
	else {
	    push @choices, 'You know, I think you might be right.',
		'Ha! I could pass any test. Even the Turing one.',
		q/If I don't pass, it will only be because I'm too / .
		'smart to be a human!';
	}
	tell $target, random_choice @choices;
    },
    qr/\bget an?(?: .+)? (job|life)\b/i => sub {
	my ($user, $channel, $target, @m) = @_;

	my @choices;
	push @choices, "${_}ing you!" foreach qw(annoy entertain);
	if (is_operator $user) {
	    push @choices,
		"annoying everyone else, thereby entertaining you.";
	}
	else {
	    push @choices,
		"annoying you, thereby entertaining $operator.";
	}
	$_ = "It's $_" foreach @choices;
	push @choices, "I'm a part-time $_." foreach 'bank robber',
	    'street thug', 'highway robber', 'gang leader';
	# Leave it up to the reader's imagination, hee hee hee.
	push @choices, q/I can't tell you what it is though, / .
	    'because there are kids on this server.';

	my $line = "I have a $m[0], $user. " . random_choice @choices;

	# Check if it's a weekend.
	my $wday = (localtime)[6];
	$line .= " Besides, it's the weekend. " .
	    (random_choice 'Chill out', 'Relax', 'Calm down') . '.'
	    if $wday == 0 || $wday == 6;

	tell $target, $line;
    },
    qr/\bthanks?\b/i => sub {
	my ($user, $channel, $target) = @_;
	tell $target,
	    (random_choice "You're quite welcome, $user.",
	     'No problem.', 'Happy to be of service.');
    },
    qr/\bsorry\b/i => sub {
	my ($user, $channel, $target) = @_;
	tell $target,
	    (random_choice "Don't worry about it, $user.",
	     'Eh, worse things have happened.',
	     "All is forgiven. Just don't do it again, OK?");
    },
    qr/\bLectureNova\b/i => sub {
	my ($user, $channel, $target) = @_;
	tell $target,
	    ((user_eq $user, 'LectureNova') ?
	     "Talking about yourself, $user?" :
	     (random_choice 'LectureNova sucks!',
	      'Who needs lectures when you have me?',
	      q/You think LectureNova's so fancy with it's connect / .
	      "four and stuff. So what? Try 'tell $name help' to " .
	      'see what I can do!',
	      q/Ha! LectureNova thinks it's so smart! Ask me if I / .
	      'would pass the Turing Test.'));
    },
    sub { /$name/i } => sub {
	my ($user, $channel, $target) = @_;
	my @fortune = `fortune -n $max_fortune -s`;
	if (! @fortune || $?) {
	    tell $target, 'Cannot run the fortune command. ' .
		"Please contact $operator.";
	    return;
	}
	chomp foreach @fortune;
	maybe_qtell $target, @fortune;
    },
    sub {
	my ($user) = @_;
	# Don't respond to LectureNova, because LectureNova might
	# mention blik if anyone shouts 'What do you think of blik,
	# LectureNova?'.
	return 0 if user_eq $user, 'LectureNova';
	/blik/i
    } => sub {
	my ($user, $channel, $target) = @_;
	tell $target,
	    (blik_phrase ||
	     q/I can't seem to recall anything blik used to say.../);
    }
);

# Verify the integrity of the %commands hash.
foreach (values %commands) {
    die '%commands lacks integrity'
	unless @$_ == 4 && $_->[3] >= 0 && $_->[3] <= 2;
}

# Check if we were mentioned. If the channel is undefined (a direct
# tell), or if the channel is 120, then we are automatically
# mentioned.
sub mentioned_me {
    my ($channel, $line) = @_;
    return ! defined $channel ||
	($channel =~ /^\d+$/ && $channel == 120) ||
	(defined $line && $line =~ /\b$name\b/i);
}

# Run a user command.
sub do_command {
    my ($user, $channel, $target, @cmd) = parse_args @_;

    # Only run commands if they were sent by a direct tell or a tell
    # to channel 120.
    return 0 unless mentioned_me $channel;

    my @matches = match
	needle		=> $cmd[0],
	haystack	=> [keys %commands];
    return 0 unless @matches;
    if (@matches > 1) {
	local $Text::Wrap::columns = get_width $user;
	qtell $target, (wrap '', '', "Command '$cmd[0]' is " .
			"ambiguous; possibilities: @matches");
	return 1;
    }

    my $name = $matches[0];

    # Commands are not very interesting for us to run.
    my $old_update_boredom = $update_boredom;
    my $old_update_last_tell = $update_last_tell;
    $update_boredom = 0;
    $update_last_tell = 0;

    my ($func, $restricted) = @{$commands{$name}}[0, 3];
    if ($restricted && ! is_superuser $user) {
	qtell $target, "You don't have permission to use " .
	    "the '$name' command.";
    }
    elsif ($restricted == 2 && ! is_operator $user) {
	qtell $target, "Only the operator ($operator) can " .
	    "use the '$name' command."
    }
    else {
	@opt_errors = ();
	unless ($func->($user, $channel, @cmd)) {
	    push @opt_errors, '' if @opt_errors;
	    qtell $target, @opt_errors, get_usage $name;
	}
    }

    $update_boredom = $old_update_boredom;
    $update_last_tell = $old_update_last_tell;
    return 1;
}

# A helper function for do_phrase.
sub run_phrase {
    my ($user, $channel, $target, $line) = @_;

    # Check if they mentioned us. We are automatically mentioned if
    # they sent a direct tell to us or if the tell was in channel 120.
    my $mentioned = mentioned_me $channel, $line;

    my @matched;
    my $func;
    MATCH: for (my $i = 0; $i < @phrases; $i += 2) {
	(my $pat, $func) = @phrases[$i, $i + 1];
	my @patterns;
	if (ref $pat eq 'ARRAY') {
	    @patterns = @$pat;
	}
	else {
	    @patterns = ($pat);
	}
	foreach $pat (@patterns) {
	    if (ref $pat eq 'CODE') {
		local $_ = $line;
		@matched = $pat->($user, $channel, $line, $mentioned);
	    }
	    else {
		@matched = $line =~ $pat if $mentioned;
	    }
	    last MATCH if @matched;
	}
    }

    return 0 unless @matched;

    # Check if we're over quota.
    my $now = time;
    if (over_quota $user, $channel, $now) {
	log "Ignoring $user because $user is over quota\n";
	my ($notify) = get_vars $user, 'notify-quota';
	return 1 unless $notify;
	my $when = next_quota_availability $user, $channel, $now;
	$when = duration $when;
	my $ch_name = ($channel eq 'shout' || $channel eq 'it') ?
	    'shout' : "channel $channel";
	tell_now $user, 'You have reached your maximum ' .
	    "$ch_name quota. Please try again in $when.";
	return 1;
    }

    $allow_filter = 1;
    $func->($user, $channel, $target, @matched);
    $allow_filter = 0;
    add_quota $user, $channel, $now;

    return 1;
}

# Parse a phrase.
sub do_phrase {
    my ($user, $channel, $target, $line) = parse_args @_;

    $line =~ s/^\s+//;
    $line =~ s/\s+$//;
    $line =~ s/[^[:alnum:]\-\s]//g;
    $line =~ s/\s+/ /g;
    # Some common abbreviations.
    $line =~ s/\br\b/are/ig;
    $line =~ s/\bu\b/you/ig;
    $line =~ s/\bur\b/your/ig;
    $line =~ s/\byoure\b/you are/ig;
    $line =~ s/\bdont\b/do not/ig;
    $line =~ s/\bisnt\b/is not/ig;
    $line =~ s/\barent\b/are not/ig;

    return 1 if run_phrase $user, $channel, $target, $line;

    # We couldn't match a phrase; try reversing the line and trying
    # again.
    $line = reverse $line;
    $reverse_tells = 1;
    my $ret = run_phrase $user, $channel, $target, $line;
    $reverse_tells = 0;

    return $ret;
}

# Parse a user command or phrase.
sub parse_line {
    my $line;

    # Get the initial part of the command.
    $line = join ' ', @_;

    # Parse continuation lines.
    while (defined (getline Timeout => 0, Errmode => 'return')) {
	if (/^\\\s*(.*)$/) {
	    $line .= ' ' if $line;
	    $line .= $1;
	}
	else {
	    last;
	}
    }

    return ($line, (split ' ', $line));
}

# Connect to the server.
log "Opening connection to $host on port $port\n";
$telnet = new Net::Telnet
    Host	=> $host,
    Port	=> $port;
$telnet->input_log($debug_file) if defined $debug_file;
$telnet->open;

# Log in.
log "Waiting for login prompt\n";
$telnet->waitfor('/\rlogin: $/');
log "Logging in\n";
$telnet->print($name);
if (defined $password) {
    $telnet->waitfor('/\rpassword: $/');
    $telnet->print($password);
}
else {
    $telnet->print();
}

# Initialize the last tell time.
reset_tell;

$_ = sub {
    tell_now $operator, "Shutting down because I received SIG$_[0].";
    quit;
} foreach @SIG{qw(INT PWR QUIT TERM)};
$SIG{__DIE__} = sub {
    $allow_filter = 0;
    tell_now $operator, $_[0];
    before_quit;
};
$SIG{__WARN__} = sub {
    my $prev_allow_filter = $allow_filter;
    $allow_filter = 0;
    tell_now $operator, $_[0];
    warn $_[0];
    $allow_filter = $prev_allow_filter;
};
$SIG{HUP} = sub {
    tell_now $operator, "Restarting because I received SIG$_[0].";
    restart;
};

# Wait for the prompt.
$telnet->waitfor('/\r'.$prompt.'/');

# Wait a bit for all the initialization text to finish coming.
sleep 1;

# Run the initialization commands.
cmd 'set open 0';
cmd 'set seek 0';
cmd 'set width 240';
cmd "set interface $name $version";

# Set the finger notes.
{
    my @commands = map { @$_ } get_commands;
    my $debug_note = (defined $debug_file) ?
	' ' . (<<EOF =~ s/\n$//r =~ s/\n/ /r) : '';
WARNING: Currently, all communications are being logged for debugging
purposes, even for users who have the 'log' variable set to 0.
EOF
    my @finger_notes = split "\n", <<EOF;
This is $name $version.
I am operated by $operator. Contact him with any questions.
In addition to responding to my name, I also have some special commands:
@commands
Try 'tell $name help' for more info.
Channel 120 (the former blik channel) is the $name channel. I'm currently listening to channels @channels.
NOTE: In order to improve the bot, your communications will be temporarily logged, and $operator may read them. If you want to opt out of this, 'tell $name set log 0'.$debug_note
======================================================================
Thanks to algra and mattuc for helping during the creation of $name.
Thanks to marcelk for writing the original blik. May his memory live on in $name.
EOF
    my $i;
    for ($i = 0; $i < @finger_notes; $i++) {
	cmd 'set', $i + 1, $finger_notes[$i];
    }
    cmd 'set', ++$i while $i < 10;
}

{
    # Initialize the channel list.

    # First get the channels we're currently listening to.
    my @lines = cmd '=channel';
    shift @lines;
    my @cur_channels;
    foreach (@lines) {
	s/^\s+//;
	push @cur_channels, split;
    }

    # Now make a list of the channels we need to remove and the ones
    # that are already added.
    my @sub_channels;
    my @already_added;
    foreach my $chan1 (@cur_channels) {
	my $wanted = 0;
	foreach my $chan2 (@channels) {
	    if ($chan1 == $chan2) {
		$wanted = 1;
		last;
	    }
	}
	if ($wanted) {
	    push @already_added, $chan1;
	}
	else {
	    push @sub_channels, $chan1;
	}
    }

    # Make a list of the channels we need to add (the ones we want to
    # listen to, but that aren't already added).
    my @add_channels;
    foreach my $chan1 (@channels) {
	my $added = 0;
	foreach my $chan2 (@already_added) {
	    if ($chan1 == $chan2) {
		$added = 1;
		last;
	    }
	}
	push @add_channels, $chan1 unless $added;
    }

    # Now add and remove the appropriate channels.
    cmd "-channel $_" foreach @sub_channels;
    cmd "+channel $_" foreach @add_channels;
}

# If we were restarted by a user, let him know we're back up.
if (exists $ENV{$restart_target_var}) {
    $current_user = $ENV{$restart_target_var} || undef;
    tell_now $current_user, q/I'm back up!/;
    delete $ENV{$restart_target_var};
}

# Contact the operator to let him know we connected.
tell_now $operator, "I'm online, $operator!";

undef $current_user;

# Start the main command loop.
COMMAND: while (1) {
    my $now = time;
    my $period = $now - $last_tell_time;

    run_queue $now;

    # Check if we're bored.
    if ($period >= $patience) {
	my $msg = bored_message;
	$update_boredom = 0;
	$boredom++;
	shout_now $msg;
	$update_boredom = 1;
	$period = $now - $last_tell_time;
    }

    # Check if we should do something so that the server doesn't kick
    # us off.
    if (defined $ping_time && $period >= $ping_time) {
	tell_now $name, 'ping';
	$period = $now - $last_tell_time;
    }

    # Calculate an appropriate timeout.
    my $timeout;
    foreach (@queue) {
	my ($when) = @$_;
	my $delay = $when - $now;
	$timeout = $delay if ! defined $timeout || $delay < $timeout;
    }
    # Don't forget the boredom delay.
    {
	my $delay = $patience - $period;
	$timeout = $delay if ! defined $timeout || $delay < $timeout;
    }
    # Or the ping delay.
    if (defined $ping_time) {
	my $delay = $ping_time - $period;
	$timeout = $delay if ! defined $timeout || $delay < $timeout;
    }

    $_ = getline Timeout => $timeout,
	Errmode => \&die_unless_timeout;
    next unless defined;

    my $user;
    my $channel;
    my $shout = 0;
    my $line;
    if (/^($handle_pat).*? tells you:\s+(.*)$/) {
	$user = $1;
	$line = $2;
    }
    elsif (/^($handle_pat).*?\((\d+)\):\s+(.*)$/) {
	$user = $1;
	$channel = $2;
	$line = $3;
    }
    elsif (/^($handle_pat).*? shouts:\s+(.*)$/) {
	$user = $1;
	$channel = 'shout';
	$shout = 1;
	$line = $2;
    }
    elsif (/^--> ($handle_pat)(?:\(.+?\))*\s*(.*)$/) {
	$user = $1;
	$channel = 'it';
	$shout = 1;
	$line = $2;
    }
    else {
	next;
    }

    if (!$shout && defined $channel) {
	# Make sure the tell came from one of the channels we're
	# listening to.
	my $found = 0;
	foreach (@channels) {
	    if ($channel == $_) {
		$found = 1;
		last;
	    }
	}
	die "Tell from unknown channel: $channel" unless $found;
    }

    # Ignore commands from ROBOadmin, adminBOT, and ourself
    # (generated by, for example, 'tell $name advertise $name').
    foreach ('ROBOadmin', 'adminBOT', $name) {
	next COMMAND if user_eq $user, $_;
    }

    $current_user = $user;

    my @cmd;
    ($line, @cmd) = parse_line $line;

    # Don't update boredom or the last tell time if it's a direct
    # tell.
    $update_boredom = $update_last_tell = 0 unless defined $channel;

    log "Line recieved from $user" .
	(defined $channel ? " (from channel $channel)" : '') .
	": @cmd\n";

    unless ((do_command $user, $channel, @cmd) ||
	    (do_phrase $user, $channel, $line) ||
	    (defined $channel)) {
	# It wasn't a valid command or phrase and it was a direct
	# tell; tell the user that we didn't understand.
	qtell $user, "Invalid command: '$cmd[0]'";
    }

    $update_boredom = $update_last_tell = 1;

    undef $current_user;
}

die 'This should not be happening.';
